@@ -15,6 +15,7 @@ mod cli {
                                        printf '%s' "${#REPLY}";"#;
     const PRINT_ARGS: &str = r#"args=("$@"); declare -p args;"#;
     const EXIT_WITH: &str = r#"exit "${1:?}";"#;
+    const EXIT_WITH_ENV: &str = r#"exit "${EXIT_WITH:?}";"#;
     const AWAIT_AND_TOUCH: &str = r#"echo awaiting; \
                                      until [[ -e "${1:?}" ]]; do sleep .1; done; \
                                      echo > "${2:?}";"#;
@@ -57,21 +58,37 @@ mod cli {
 
     fn succeed(cmd: &mut Command) -> String {
         let result = run(cmd);
-        assert_eq!(result.err, "");
+        if cfg!(feature="debug") {
+            if !result.err.is_empty() { eprintln!("stderr:\n{}", result.err); }
+        } else {
+            // debug writes to stderr, so don't bother checking it in that mode
+            assert_eq!(result.err, "");
+        }
         assert_eq!(result.status, Some(0));
         result.out
     }
 
-    fn modtime<P: AsRef<Path>>(path: P) -> SystemTime {
-        std::fs::metadata(path).expect("No metadata").modified().expect("No modtime")
+    // Returns once the given file contains different contents than those provided. Panics if the
+    // file does not change after ~5s.
+    //
+    // Note this could return immediately if the file already doesn't contain initial_contents
+    // (e.g. if the given contents were wrong) because such a check could race. Do additional
+    // checks prior to waiting if needed.
+    fn wait_for_contents_to_change<P: AsRef<Path>>(file: P, initial_contents: &str) {
+        for _ in 1..50 {
+            if std::fs::read_to_string(&file).unwrap() != initial_contents { return; }
+            std::thread::sleep(Duration::from_millis(100));
+        }
+        panic!("Contents of {} did not change", file.as_ref().to_string_lossy());
     }
 
     fn make_dir_stale<P: AsRef<Path>>(dir: P, age: Duration) -> Result<()> {
+        debug_assert!(dir.as_ref().is_dir());
         let desired_time = SystemTime::now() - age;
         let stale_time = filetime::FileTime::from_system_time(desired_time);
         for entry in std::fs::read_dir(dir)? {
             let path = entry?.path();
-            let last_modified = modtime(&path);
+            let last_modified = std::fs::metadata(&path)?.modified()?;
 
             if path.is_file() && last_modified > desired_time {
                 filetime::set_file_mtime(&path, stale_time)?;
@@ -82,6 +99,14 @@ mod cli {
         Ok(())
     }
 
+    fn make_file_stale<P: AsRef<Path>>(file: P, age: Duration) -> Result<()> {
+        debug_assert!(file.as_ref().is_file());
+        let desired_time = SystemTime::now() - age;
+        let stale_time = filetime::FileTime::from_system_time(desired_time);
+        filetime::set_file_mtime(&file, stale_time)?;
+        Ok(())
+    }
+
     fn join<A: Clone>(beg: &[A], tail: &[A]) -> Vec<A> {
         beg.iter().chain(tail).cloned().collect()
     }
@@ -90,7 +115,7 @@ mod cli {
     fn help() {
         let dir = TestDir::temp();
         let out = succeed(bkt(dir.path("cache")).arg("--help"));
-        assert!(out.contains("bkt [OPTIONS] [--] <COMMAND>..."));
+        assert!(out.contains("bkt [OPTIONS] -- <COMMAND>..."));
     }
 
     #[test]
@@ -102,7 +127,12 @@ mod cli {
 
         for _ in 1..3 {
             let subsequent_result = run(bkt(dir.path("cache")).args(args));
-            assert_eq!(first_result, subsequent_result);
+            if cfg!(feature="debug") {
+                assert_eq!(first_result.status, subsequent_result.status);
+                assert_eq!(first_result.out, subsequent_result.out);
+            } else {
+                assert_eq!(first_result, subsequent_result);
+            }
         }
     }
 
@@ -114,12 +144,19 @@ mod cli {
         let first_result = succeed(bkt(dir.path("cache")).args(args));
         assert_eq!(first_result, "1");
 
+        // Slightly stale is still cached
+        make_dir_stale(dir.path("cache"), Duration::from_secs(10)).unwrap();
         let subsequent_result = succeed(bkt(dir.path("cache")).args(args));
         assert_eq!(first_result, subsequent_result);
 
         make_dir_stale(dir.path("cache"), Duration::from_secs(120)).unwrap();
         let after_stale_result = succeed(bkt(dir.path("cache")).args(args));
         assert_eq!(after_stale_result, "2");
+
+        // Respects BKT_TTL env var (other tests cover --ttl)
+        make_dir_stale(dir.path("cache"), Duration::from_secs(10)).unwrap();
+        let env_result = succeed(bkt(dir.path("cache")).env("BKT_TTL", "5s").args(args));
+        assert_eq!(env_result, "3");
     }
 
     #[test]
@@ -172,15 +209,10 @@ mod cli {
         let args = ["--stale=10s", "--ttl=20s", "--", "bash", "-c", COUNT_INVOCATIONS, "arg0", file.to_str().unwrap()];
         assert_eq!(succeed(bkt(dir.path("cache")).args(args)), "1");
 
-        let last_mod = modtime(&file);
         make_dir_stale(dir.path("cache"), Duration::from_secs(15)).unwrap();
         assert_eq!(succeed(bkt(dir.path("cache")).args(args)), "1");
 
-        for _ in 1..10 {
-            if modtime(&file) > last_mod { break; }
-            std::thread::sleep(Duration::from_millis(100));
-        }
-        assert!(modtime(&file) > last_mod);
+        wait_for_contents_to_change(&file, ".");
         assert_eq!(std::fs::read_to_string(&file).unwrap(), "..");
         assert_eq!(succeed(bkt(dir.path("cache")).args(args)), "2");
     }
@@ -192,44 +224,55 @@ mod cli {
         let cmd = format!("{} false;", COUNT_INVOCATIONS);
         let args = ["--discard-failures", "--", "bash", "-c", &cmd, "arg0", file.to_str().unwrap()];
 
-        assert_eq!(run(bkt(dir.path("cache")).args(args)),
-                   CmdResult { out: "1".into(), err: "".into(), status: Some(1) });
+        let result = run(bkt(dir.path("cache")).args(args));
+        assert_eq!(result.out, "1");
+        assert_eq!(result.status, Some(1));
+
         // Not cached
-        assert_eq!(run(bkt(dir.path("cache")).args(args)),
-                   CmdResult { out: "2".into(), err: "".into(), status: Some(1) });
+        let result = run(bkt(dir.path("cache")).args(args));
+        assert_eq!(result.out, "2");
+        assert_eq!(result.status, Some(1));
+    }
+
+    #[test]
+    fn discard_failure_cached_separately() {
+        let dir = TestDir::temp();
+
+        let allow_args = ["--", "bash", "-c", EXIT_WITH_ENV, "arg0"];
+        let discard_args = join(&["--discard-failures"], &allow_args);
+
+        // without separate caches a --discard-failures invocation could return a previously-cached
+        // failed result. In 0.5.4 and earlier this would mean result2.status == 14.
+        let result1 = run(bkt(dir.path("cache")).args(allow_args).env("EXIT_WITH", "14"));
+        assert_eq!(result1.status, Some(14));
+        let result2 = run(bkt(dir.path("cache")).args(discard_args).env("EXIT_WITH", "0"));
+        assert_eq!(result2.status, Some(0));
     }
 
-    // Note: this test has been flaky in the past
     #[test]
     fn discard_failures_in_background() {
         let dir = TestDir::temp();
         let file = dir.path("file");
-        let cmd = format!("{} false;", COUNT_INVOCATIONS);
-        let args = ["--ttl=20s", "--", "bash", "-c", &cmd, "arg0", file.to_str().unwrap()];
-        let discard_args = join(&["--discard-failures"], &args);
-        let discard_stale_args = join(&["--stale=10s"], &discard_args);
+        let cmd = format!("{} ! \"${{FAIL:-false}}\";", COUNT_INVOCATIONS);
+        let args = ["--ttl=20s", "--discard-failures", "--", "bash", "-c", &cmd, "arg0", file.to_str().unwrap()];
+        let stale_args = join(&["--stale=10s"], &args);
 
-        // cached
-        assert_eq!(run(bkt(dir.path("cache")).args(args)),
-                   CmdResult { out: "1".into(), err: "".into(), status: Some(1) });
+        // Cache result normally
+        assert_eq!(succeed(bkt(dir.path("cache")).args(args)), "1");
+
+        // Cause cmd to fail and not be cached
+        std::env::set_var("FAIL", "true");
 
         // returns cached result, but attempts to warm in the background
-        let last_mod = modtime(&file);
         make_dir_stale(dir.path("cache"), Duration::from_secs(15)).unwrap();
-        assert_eq!(run(bkt(dir.path("cache")).args(&discard_stale_args)),
-                   CmdResult { out: "1".into(), err: "".into(), status: Some(1) });
+        assert_eq!(succeed(bkt(dir.path("cache")).args(&stale_args)), "1");
 
-        for _ in 1..10 {
-            if modtime(&file) > last_mod { break; }
-            std::thread::sleep(Duration::from_millis(100));
-        }
-        // Command ran
-        assert!(modtime(&file) > last_mod, "{:?} !> {:?}", modtime(&file), last_mod);
+        // Verify command ran
+        wait_for_contents_to_change(&file, ".");
         assert_eq!(std::fs::read_to_string(&file).unwrap(), "..");
 
-        // but cache was not updated
-        assert_eq!(run(bkt(dir.path("cache")).args(&discard_args)),
-                   CmdResult { out: "1".into(), err: "".into(), status: Some(1) });
+        // But cached success is still returned
+        assert_eq!(succeed(bkt(dir.path("cache")).args(args)), "1");
     }
 
     #[test]
@@ -244,6 +287,9 @@ mod cli {
 
         let diff_cache = succeed(bkt(dir.path("cache")).arg(format!("--cache-dir={}", dir.path("new-cache").display())).args(args));
         assert_eq!(diff_cache, "2");
+
+        let env_cache = succeed(bkt(dir.path("cache")).env("BKT_CACHE_DIR", dir.path("env-cache").as_os_str()).args(args));
+        assert_eq!(env_cache, "3");
     }
 
     // https://github.com/dimo414/bkt/issues/9
@@ -275,6 +321,8 @@ mod cli {
         assert_eq!(diff_scope, "2");
         assert_eq!(diff_scope, succeed(bkt(dir.path("cache"))
             .arg("--scope=foo").args(args)));
+        assert_eq!(diff_scope, succeed(bkt(dir.path("cache"))
+            .env("BKT_SCOPE", "foo").args(args)));
     }
 
     #[test]
@@ -341,6 +389,31 @@ mod cli {
     }
 
     #[test]
+    fn respects_modtime() {
+        let dir = TestDir::temp();
+        let file = dir.path("file");
+        let watch_file = dir.path("watch");
+        let args = ["--modtime", watch_file.to_str().unwrap(), "--", "bash", "-c", COUNT_INVOCATIONS, "arg0", file.to_str().unwrap()];
+        let no_file_result = succeed(bkt(dir.path("cache")).args(args));
+        // File absent is cached
+        assert_eq!(no_file_result, "1");
+        assert_eq!(no_file_result, succeed(bkt(dir.path("cache")).args(args)));
+
+        // create a new file, invalidating cache
+        File::create(&watch_file).unwrap();
+        let new_file_result = succeed(bkt(dir.path("cache")).args(args));
+        assert_eq!(new_file_result, "2");
+        assert_eq!(new_file_result, succeed(bkt(dir.path("cache")).args(args)));
+
+        // update the modtime, again invalidating the cache
+        make_file_stale(&watch_file, Duration::from_secs(10)).unwrap();
+        let old_file_result = succeed(bkt(dir.path("cache")).args(args));
+        assert_eq!(old_file_result, "3");
+        assert_eq!(old_file_result, succeed(bkt(dir.path("cache")).args(args)));
+    }
+
+    #[test]
+    #[cfg(not(feature="debug"))]
     fn no_debug_output() {
         let dir = TestDir::temp();
         let args = ["--", "bash", "-c", "true"];
@@ -353,6 +426,27 @@ mod cli {
                    CmdResult { out: "".into(), err: "".into(), status: Some(0) });
     }
 
+    #[test]
+    #[cfg(feature="debug")]
+    fn debug_output() {
+        fn starts_with_bkt(s: &str) -> bool { s.lines().all(|l| l.starts_with("bkt: ")) }
+
+        let miss_debug_re = regex::Regex::new(
+            "bkt: state: \nbkt: lookup .* not found\nbkt: cleanup data .*\nbkt: cleanup keys .*\nbkt: store data .*\nbkt: store key .*\n").unwrap();
+        let hit_debug_re = regex::Regex::new("bkt: lookup .* found\n").unwrap();
+
+        let dir = TestDir::temp();
+        let args = ["--", "bash", "-c", PRINT_ARGS, "arg0"];
+
+        let miss = run(bkt(dir.path("cache")).args(args));
+        assert!(starts_with_bkt(&miss.err), "{}", miss.err);
+        assert!(miss_debug_re.is_match(&miss.err), "{}", miss.err);
+
+        let hit = run(bkt(dir.path("cache")).args(args));
+        assert!(starts_with_bkt(&hit.err), "{}", hit.err);
+        assert!(hit_debug_re.is_match(&hit.err), "{}", hit.err);
+    }
+
     #[test]
     fn output_preserved() {
         let dir = TestDir::temp();
@@ -372,6 +466,7 @@ mod cli {
     }
 
     #[test]
+    #[cfg(not(feature="debug"))]
     fn sensitive_output() {
         let dir = TestDir::temp();
         let args = ["--", "bash", "-c", r"printf 'foo\0bar'; printf 'bar\0baz\n' >&2"];
@@ -450,10 +545,10 @@ mod cli {
         let proc1 = bkt(dir.path("cache")).args(args).stdout(Stdio::piped()).stderr(Stdio::piped()).spawn().unwrap();
         let proc2 = bkt(dir.path("cache")).args(args).stdout(Stdio::piped()).stderr(Stdio::piped()).spawn().unwrap();
         let result1: CmdResult = proc1.wait_with_output().unwrap().into();
-        assert_eq!(result1.err, "");
+        if !cfg!(feature="debug") { assert_eq!(result1.err, ""); }
         assert_eq!(result1.status, Some(0));
         let result2: CmdResult = proc2.wait_with_output().unwrap().into();
-        assert_eq!(result2.err, "");
+        if !cfg!(feature="debug") { assert_eq!(result2.err, ""); }
         assert_eq!(result2.status, Some(0));
 
         assert_eq!(std::fs::read_to_string(&file).unwrap(), "..");
